package org.apache.bcel.classfile;

import java.util.Map;
import java.util.LinkedHashMap;
import java.util.HashMap;
import java.io.DataOutputStream;
import java.util.Objects;
import java.io.IOException;
import java.io.DataInput;

public final class ConstantUtf8 extends Constant
{
    private static volatile int considered;
    private static volatile int created;
    private static volatile int hits;
    private static volatile int skipped;
    private static final String SYS_PROP_CACHE_MAX_ENTRIES = "bcel.maxcached";
    private static final String SYS_PROP_CACHE_MAX_ENTRY_SIZE = "bcel.maxcached.size";
    private static final String SYS_PROP_STATISTICS = "bcel.statistics";
    private final String value;
    
    public static synchronized void clearCache() {
        Cache.CACHE.clear();
    }
    
    static synchronized void clearStats() {
        ConstantUtf8.hits = (ConstantUtf8.considered = (ConstantUtf8.skipped = (ConstantUtf8.created = 0)));
    }
    
    public static ConstantUtf8 getCachedInstance(final String value) {
        if (value.length() > Cache.MAX_ENTRY_SIZE) {
            ++ConstantUtf8.skipped;
            return new ConstantUtf8(value);
        }
        ++ConstantUtf8.considered;
        synchronized (ConstantUtf8.class) {
            ConstantUtf8 result = Cache.CACHE.get(value);
            if (result != null) {
                ++ConstantUtf8.hits;
                return result;
            }
            result = new ConstantUtf8(value);
            Cache.CACHE.put(value, result);
            return result;
        }
    }
    
    public static ConstantUtf8 getInstance(final DataInput dataInput) throws IOException {
        return getInstance(dataInput.readUTF());
    }
    
    public static ConstantUtf8 getInstance(final String value) {
        return Cache.isEnabled() ? getCachedInstance(value) : new ConstantUtf8(value);
    }
    
    static void printStats() {
        final String prefix = "[Apache Commons BCEL]";
        System.err.printf("%s Cache hit %,d/%,d, %d skipped.%n", "[Apache Commons BCEL]", ConstantUtf8.hits, ConstantUtf8.considered, ConstantUtf8.skipped);
        System.err.printf("%s Total of %,d ConstantUtf8 objects created.%n", "[Apache Commons BCEL]", ConstantUtf8.created);
        System.err.printf("%s Configuration: %s=%,d, %s=%,d.%n", "[Apache Commons BCEL]", "bcel.maxcached", Cache.MAX_ENTRIES, "bcel.maxcached.size", Cache.MAX_ENTRY_SIZE);
    }
    
    public ConstantUtf8(final ConstantUtf8 constantUtf8) {
        this(constantUtf8.getBytes());
    }
    
    ConstantUtf8(final DataInput dataInput) throws IOException {
        super((byte)1);
        this.value = dataInput.readUTF();
        ++ConstantUtf8.created;
    }
    
    public ConstantUtf8(final String value) {
        super((byte)1);
        this.value = Objects.requireNonNull(value, "value");
        ++ConstantUtf8.created;
    }
    
    @Override
    public void accept(final Visitor v) {
        v.visitConstantUtf8(this);
    }
    
    @Override
    public void dump(final DataOutputStream file) throws IOException {
        file.writeByte(super.getTag());
        file.writeUTF(this.value);
    }
    
    public String getBytes() {
        return this.value;
    }
    
    @Deprecated
    public void setBytes(final String bytes) {
        throw new UnsupportedOperationException();
    }
    
    @Override
    public String toString() {
        return super.toString() + "(\"" + Utility.replace(this.value, "\n", "\\n") + "\")";
    }
    
    static {
        if (Cache.BCEL_STATISTICS) {
            Runtime.getRuntime().addShutdownHook(new Thread(ConstantUtf8::printStats));
        }
    }
    
    private static class Cache
    {
        private static final boolean BCEL_STATISTICS;
        private static final int MAX_ENTRIES;
        private static final int INITIAL_CAPACITY;
        private static final HashMap<String, ConstantUtf8> CACHE;
        private static final int MAX_ENTRY_SIZE;
        
        static boolean isEnabled() {
            return Cache.MAX_ENTRIES > 0 && Cache.MAX_ENTRY_SIZE > 0;
        }
        
        static {
            BCEL_STATISTICS = Boolean.getBoolean("bcel.statistics");
            MAX_ENTRIES = Integer.getInteger("bcel.maxcached", 0);
            INITIAL_CAPACITY = (int)(Cache.MAX_ENTRIES / 0.75);
            CACHE = new LinkedHashMap<String, ConstantUtf8>(Cache.INITIAL_CAPACITY, 0.75f, true) {
                private static final long serialVersionUID = -8506975356158971766L;
                
                @Override
                protected boolean removeEldestEntry(final Map.Entry<String, ConstantUtf8> eldest) {
                    return this.size() > Cache.MAX_ENTRIES;
                }
            };
            MAX_ENTRY_SIZE = Integer.getInteger("bcel.maxcached.size", 200);
        }
    }
}
